#include <cassert>
#include <cstdlib>
#include <ctype.h>
#include <fcntl.h>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include "opencl.hpp"

const char* cmdname;

void usage() {
   std::cerr << "Usage: " << cmdname << " source.cl" << std::endl;
   exit(1);
}

void print_binary_object(std::ostream& out, const std::string& name,
      unsigned char* binary, size_t binary_size) {
   out << "static const size_t " << name << "_size = " << binary_size << ";" << std::endl;
   out << "static const unsigned char " << name << "[] = {\n";
   bool start = true; bool len = 0; bool newline = false;
   for (unsigned int i = 0; i < binary_size; ++i) {
      if (len >= 60) {
	 out << "\"\n";
	 start = true;
      }
      if (start) {
	 if (newline) {
	    out << "\n"; newline = false;
	 }
	 out << "   \"";
	 start = false; len = 0;
      }
      unsigned int ch = binary[i];
      if (isprint(ch) && ch != '"') {
	 out << (char) ch; ++len;
      } else if (ch == '\t') {
	 out << "\\t"; len += 2;
      } else if (ch == '\n') {
	 out << "\\n\""; start = true; newline = true;
      } else {
	 out << "\\" << std::oct << std::setfill('0') << std::setw(3) << ch;
	 len += 4;
      }
   }
   if (!newline) {
      out << "\"";
   }
   out << "\n};\n";
}

int main(int argc, char** argv) {
   /* process command line arguments */
   cmdname = *argv++; --argc;
   if (argc != 1) usage();
   const char* source = *argv++; --argc;
   std::string source_string(source);
   size_t source_string_len = source_string.size();
   if (source_string_len <= 3 ||
	 source_string.substr(source_string_len - 3, std::string::npos) != ".cl") {
      std::cerr << cmdname << ": source file should end in ``.cl''" << std::endl;
      exit(1);
   }
   std::string stem = source_string.substr(0, source_string_len-3);

   /* map source file into our address space */
   int fd = open(source, O_RDONLY);
   if (fd < 0) perror(source), exit(1);
   struct stat statbuf;
   if (fstat(fd, &statbuf) < 0) perror(source), exit(1);
   size_t source_nbytes = statbuf.st_size;
   if (source_nbytes == 0) {
      std::cerr << cmdname << ": empty source file: " << source << std::endl;
   }
   const char* source_buf = (char*) mmap(0, source_nbytes, PROT_READ, MAP_SHARED, fd, 0);
   if (source_buf == MAP_FAILED) perror(source), exit(1);

   /* open context for the first GPU device that we find */
   cl_platform_id platform;
   CHECK_OPENCL(clGetPlatformIDs, 1, &platform, 0);
   cl_device_id device_id;
   CHECK_OPENCL(clGetDeviceIDs, platform, CL_DEVICE_TYPE_GPU, 1, &device_id, 0);
   cl_int errcode = 0;
   cl_context context = clCreateContext(/* properties = */ 0,
      /* num_devices = */ 1, /* devices = */ &device_id, /* pfn_notify = */ 0,
      /* user_data = */ 0, &errcode);
   CHECK_OPENCL_ERRCODE("clCreateContext", errcode);

   cl_program program = clCreateProgramWithSource(context, 1,
      &source_buf, &source_nbytes, &errcode);
   CHECK_OPENCL_ERRCODE("clCreateProgramWithSource", errcode);

   errcode = clBuildProgram(program, 1, &device_id,
      /* options = */ 0, /* pfn_notify = */ 0, /* user_data = */ 0);
   if (errcode == CL_BUILD_PROGRAM_FAILURE) {
      size_t log_len;
      CHECK_OPENCL(clGetProgramBuildInfo, program, device_id,
	 CL_PROGRAM_BUILD_LOG, 0, 0, &log_len);
      char errlog[log_len];
      CHECK_OPENCL(clGetProgramBuildInfo, program, device_id,
	 CL_PROGRAM_BUILD_LOG, log_len, errlog, 0);
      std::cerr << errlog; // is nullbyte-terminated
      exit(1);
   } else {
      CHECK_OPENCL_ERRCODE("clBuildProgram", errcode);
   }

   /* retrieve binary representation */
   size_t number_of_devices = 0;
   CHECK_OPENCL(clGetProgramInfo, program, CL_PROGRAM_NUM_DEVICES,
      sizeof number_of_devices, &number_of_devices, 0);
   assert(number_of_devices == 1); // we ignore all other devices
   size_t binary_size[number_of_devices];
   CHECK_OPENCL(clGetProgramInfo, program, CL_PROGRAM_BINARY_SIZES,
      sizeof binary_size, binary_size, 0);
   unsigned char* binaries[number_of_devices];
   for (unsigned int i = 0; i < number_of_devices; ++i) {
      binaries[i] = new unsigned char[binary_size[i]];
   }
   CHECK_OPENCL(clGetProgramInfo, program, CL_PROGRAM_BINARIES,
      number_of_devices * sizeof(unsigned char*), binaries, 0);

   std::string output_file(stem); output_file += ".hpp";
   std::ofstream out(output_file.c_str());
   if (!out) {
      perror(output_file.c_str()); exit(1);
   }
   for (unsigned int i = 0; i < stem.size(); ++i) {
      if (stem[i] == '-') stem[i] = '_';
   }
   print_binary_object(out, stem, binaries[0], binary_size[0]);
}
